from __future__ import annotations

from typing import Any, Callable, Mapping

from lilya import status
from lilya._internal._exception_handlers import (
    ExceptionHandlers,
    StatusHandlers,
    wrap_app_handling_exceptions,
)
from lilya.enums import ScopeType
from lilya.exceptions import HTTPException, WebSocketException
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.requests import Request
from lilya.responses import PlainText, Response
from lilya.types import ASGIApp, Receive, Scope, Send
from lilya.websockets import WebSocket


class ExceptionMiddleware(MiddlewareProtocol):
    """
    Middleware for handling exceptions in an ASGI application.

    This middleware allows registering custom exception handlers and status code handlers.

    Args:
        app (ASGIApp): The ASGI application.
        handlers (Optional[Mapping]): Custom exception handlers.
        debug (bool): Enable debug mode for handling 404 cases.

    Usage:
    ```python
    app = ExceptionMiddleware(app, handlers={404: custom_handler}, debug=True)
    ```

    """

    def __init__(
        self,
        app: ASGIApp,
        handlers: Mapping[Any, Callable[[Request, Exception], Response]] | None = None,
        debug: bool = False,
    ) -> None:
        self.app = app
        self.debug = debug
        self._status_handlers: StatusHandlers = {}
        self._exception_handlers: ExceptionHandlers = {
            HTTPException: self.http_exception,
            WebSocketException: self.websocket_exception,
        }
        if handlers is not None:
            self._add_handlers(handlers)

    def _add_handlers(
        self, handlers: Mapping[Any, Callable[[Request, Exception], Response]]
    ) -> None:
        for key, value in handlers.items():
            if isinstance(key, int):
                self._status_handlers[key] = value
            else:
                assert issubclass(key, Exception)
                self._exception_handlers[key] = value

    def add_exception_handler(
        self,
        exception_or_status: type[Exception] | int,
        handler: Callable[[Request, Exception], Response],
    ) -> None:
        """
        Add a custom exception handler.

        Args:
            exception_or_status (Union[Type[Exception], int]): Exception class or status code.
            handler (Callable[[Request, Exception], Response]): Exception handler function.

        Usage:
        ```python
        app.add_exception_handler(404, custom_handler)
        ```

        """
        if isinstance(exception_or_status, int):
            self._status_handlers[exception_or_status] = handler
        else:
            assert issubclass(exception_or_status, Exception)
            self._exception_handlers[exception_or_status] = handler

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle incoming ASGI scope.

        Args:
            scope (Scope): ASGI scope.
            receive (Receive): ASGI receive function.
            send (Send): ASGI send function.

        """
        if scope["type"] not in (ScopeType.HTTP, ScopeType.WEBSOCKET):
            await self.app(scope, receive, send)
            return

        scope["lilya.exception_handlers"] = (
            self._exception_handlers,
            self._status_handlers,
        )

        connection = self._get_connection(scope, receive, send)
        await wrap_app_handling_exceptions(self.app, connection)(scope, receive, send)

    def _get_connection(self, scope: Scope, receive: Receive, send: Send) -> Request | WebSocket:
        """
        Get the appropriate connection object based on the ASGI scope type.

        Args:
            scope (Scope): ASGI scope.

        Returns:
            Union[Request, WebSocket]: Connection object.

        """
        if scope["type"] == ScopeType.HTTP:
            return Request(scope, receive, send)
        return WebSocket(scope, receive, send)

    def http_exception(self, request: Request, exc: Exception) -> Response:
        """
        Handle HTTP exceptions.

        Args:
            request (Request): HTTP request.
            exc (Exception): HTTP exception.

        Returns:
            Response: HTTP response.

        """
        assert isinstance(exc, HTTPException)
        if exc.status_code in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}:
            return Response(status_code=exc.status_code, headers=exc.headers)
        return PlainText(exc.detail, status_code=exc.status_code, headers=exc.headers)

    async def websocket_exception(self, websocket: WebSocket, exc: Exception) -> None:
        """
        Handle WebSocket exceptions.

        Args:
            websocket (WebSocket): WebSocket connection.
            exc (Exception): WebSocket exception.

        """
        assert isinstance(exc, WebSocketException)
        await websocket.close(code=exc.code, reason=exc.reason)



import click

from lilya.cli.directives.operations._constants import SECRET_KEY_INSECURE_PREFIX
from lilya.cli.exceptions import DirectiveError
from lilya.cli.templates import TemplateDirective
from lilya.cli.terminal import Print
from lilya.crypto import get_random_secret_key

printer = Print()


@click.option(
    "--with-deployment",
    is_flag=True,
    show_default=True,
    default=False,
    help="Creates a project with base deployment files.",
)
@click.option(
    "--deployment-folder-name",
    default="deployment",
    show_default=True,
    type=str,
    help="The name of the folder for the deployment files.",
)
@click.option(
    "--with-structure",
    is_flag=True,
    show_default=True,
    default=False,
    help="Creates a project with a given structure of folders and files.",
)
@click.option("-v", "--verbosity", default=1, type=int, help="Displays the files generated.")
@click.argument("name", type=str)
@click.command(name="createproject")
def create_project(
    name: str,
    verbosity: int,
    with_deployment: bool,
    deployment_folder_name: str,
    with_structure: bool,
) -> None:
    """
    Creates the scaffold of a project.

    How to run: `lilya createproject <NAME>`

    Example: `lilya createproject myproject`
    """
    options = {
        "secret_key": SECRET_KEY_INSECURE_PREFIX + get_random_secret_key(),
        "verbosity": verbosity,
        "with_deployment": with_deployment,
        "deployment_folder_name": deployment_folder_name,
        "is_simple": with_structure,
    }
    directive = TemplateDirective()

    try:
        directive.handle("project", name=name, **options)
        printer.write_success(f"Project {name} generated successfully!")
    except DirectiveError as e:
        printer.write_error(str(e))



from __future__ import annotations

import hashlib
import hmac
import secrets
from typing import Literal

from lilya.datastructures import Cookie, Header
from lilya.enums import ScopeType
from lilya.exceptions import PermissionDenied
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.requests import Request
from lilya.types import ASGIApp, Message, Receive, Scope, Send

CSRF_SECRET_BYTES = 32
CSRF_SECRET_LENGTH = CSRF_SECRET_BYTES * 2


class CSRFMiddleware(MiddlewareProtocol):
    """
    CSRF Middleware class.

    This Middleware protects against attacks by setting a CSRF cookie with a token
    and verifying it in request headers.

    Args:
        app: The 'next' ASGI app to call.
        config: The CSRFConfig instance.
    """

    def __init__(
        self,
        app: ASGIApp,
        secret: str,
        cookie_name: str | None = None,
        header_name: str | None = None,
        cookie_path: str | None = None,
        safe_methods: set[str] | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
        domain: str | None = None,
    ) -> None:
        super().__init__(app)
        self.app = app
        self.secret = secret
        self.cookie_name = cookie_name or "csrftoken"
        self.header_name = header_name or "X-CSRFToken"
        self.cookie_path = cookie_path or "/"
        self.safe_methods = safe_methods or {"GET", "HEAD"}
        self.cookie_secure = secure
        self.cookie_httponly = httponly
        self.cookie_samesite = samesite
        self.cookie_domain = domain

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handles incoming requests, checks for CSRF token, and processes the request accordingly.

        Args:
            scope: The ASGI scope.
            receive: The ASGI receive function.
            send: The ASGI send function.
        """
        if scope["type"] != ScopeType.HTTP:
            await self.app(scope, receive, send)
            return

        request = Request(scope=scope, receive=receive, send=send)
        csrf_cookie = request.cookies.get(self.cookie_name)
        current_token = request.headers.get(self.header_name)

        if request.method in self.safe_methods:
            await self.app(scope, receive, self.get_send_wrapper(send, csrf_cookie))
        elif self._csrf_tokens_match(current_token, csrf_cookie):
            await self.app(scope, receive, send)
        else:
            raise PermissionDenied(detail="CSRF token verification failed.")

    def get_send_wrapper(self, send: Send, csrf_cookie: str | None) -> Send:
        """
        Wraps the original send function to inject a CSRF cookie if needed.

        Args:
            send: The original ASGI send function.
            csrf_cookie: The CSRF token from the request.

        Returns:
            Wrapped send function.
        """

        async def send_wrapper(message: Message) -> None:
            """
            Send function that wraps the original send to inject a cookie.

            Args:
                message: An ASGI 'Message'

            Returns:
                None
            """
            if csrf_cookie is None and message["type"] == "http.response.start":
                message.setdefault("headers", [])
                message = self._set_cookie_if_needed(message)
            await send(message)

        return send_wrapper

    def _set_cookie_if_needed(self, message: Message) -> Message:
        """
        Sets CSRF cookie in the response headers if not present.

        Args:
            message: An ASGI 'Message'
        """
        headers = Header.from_scope(scope=message)
        if "set-cookie" not in headers:
            cookie = Cookie(
                key=self.cookie_name,
                value=self._generate_csrf_token(),
                path=self.cookie_path,
                secure=self.cookie_secure,
                httponly=self.cookie_httponly,
                samesite=self.cookie_samesite,
                domain=self.cookie_domain,
            )
            headers.add("set-cookie", cookie.to_header(header=""))
            message["headers"] = headers.get_multi_items()
        return message

    def _generate_csrf_hash(self, token: str) -> str:
        """
        Generate an HMAC that signs the CSRF token.

        Args:
            token: The CSRF token.

        Returns:
            Signed HMAC of the token.
        """
        return hmac.new(self.secret.encode(), token.encode(), hashlib.sha256).hexdigest()

    def _generate_csrf_token(self) -> str:
        """
        Generate a CSRF token that includes a randomly generated string signed by an HMAC.

        Returns:
            CSRF token.
        """
        token = secrets.token_hex(CSRF_SECRET_BYTES)
        token_hash = self._generate_csrf_hash(token)
        return token + token_hash

    def _decode_csrf_token(self, token: str) -> str | None:
        """
        Decode a CSRF token and validate its HMAC.

        Args:
            token: The CSRF token.

        Returns:
            Decoded CSRF token if valid, otherwise None.
        """
        if len(token) < CSRF_SECRET_LENGTH + 1:
            return None

        token_secret = token[:CSRF_SECRET_LENGTH]
        existing_hash = token[CSRF_SECRET_LENGTH:]
        expected_hash = self._generate_csrf_hash(token_secret)
        if not secrets.compare_digest(existing_hash, expected_hash):
            return None

        return token_secret

    def _csrf_tokens_match(
        self, request_csrf_token: str | None, cookie_csrf_token: str | None
    ) -> bool:
        """
        Takes the CSRF tokens from the request and the cookie and verifies both are valid and identical.

        Args:
            request_csrf_token: CSRF token from the request headers.
            cookie_csrf_token: CSRF token from the cookie.

        Returns:
            True if tokens are valid and identical, False otherwise.
        """
        if not (request_csrf_token and cookie_csrf_token):
            return False

        decoded_request_token = self._decode_csrf_token(request_csrf_token)
        decoded_cookie_token = self._decode_csrf_token(cookie_csrf_token)
        if decoded_request_token is None or decoded_cookie_token is None:
            return False

        return secrets.compare_digest(decoded_request_token, decoded_cookie_token)



from __future__ import annotations

import select
import sys
from typing import Any, Callable, Sequence

import click
import nest_asyncio

from lilya._internal._events import AyncLifespanContextManager
from lilya.cli.directives.operations.shell.enums import ShellOption
from lilya.cli.env import DirectiveEnv
from lilya.compat import run_sync


@click.option(
    "--kernel",
    default="ipython",
    type=click.Choice(["ipython", "ptpython"]),
    help="Which shell should start.",
    show_default=True,
)
@click.command()
def shell(env: DirectiveEnv, kernel: bool) -> None:
    """
    Starts an interactive ipython shell with all the models
    and important python libraries.
    """
    if (
        sys.platform != "win32"
        and not sys.stdin.isatty()
        and select.select([sys.stdin], [], [], 0)[0]
    ):
        exec(sys.stdin.read(), globals())
        return

    on_startup = getattr(env.app, "on_startup", [])
    on_shutdown = getattr(env.app, "on_shutdown", [])
    lifespan = getattr(env.app, "lifespan", None)
    lifespan = handle_lifespan_events(
        on_startup=on_startup, on_shutdown=on_shutdown, lifespan=lifespan
    )
    run_sync(run_shell(env.app, lifespan, kernel))  # type: ignore
    return None


async def run_shell(app: Any, lifespan: Any, kernel: str) -> None:
    """Executes the database shell connection"""

    async with lifespan(app):
        if kernel == ShellOption.IPYTHON:
            from lilya.cli.directives.operations.shell.ipython import get_ipython

            ipython_shell = get_ipython(app=app)
            nest_asyncio.apply()
            ipython_shell()
        else:
            from lilya.cli.directives.operations.shell.ptpython import get_ptpython

            ptpython = get_ptpython(app=app)
            nest_asyncio.apply()
            ptpython()


def handle_lifespan_events(
    on_startup: Sequence[Callable] | None = None,
    on_shutdown: Sequence[Callable] | None = None,
    lifespan: Any | None = None,
) -> Any:
    if lifespan:
        return lifespan
    return AyncLifespanContextManager(on_startup=on_startup, on_shutdown=on_shutdown)



from enum import Enum

from dymmond_settings.enums import EnvironmentType as EnvironmentType


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value  # type: ignore

    def __repr__(self) -> str:
        return str(self)



from __future__ import annotations

from inspect import isclass
from typing import Any

from typing_extensions import get_origin


def is_class_and_subclass(value: Any, _type: Any) -> bool:
    """
    Checks if a `value` is of type class and subclass.
    by checking the origin of the value against the type being
    verified.
    """
    original = get_origin(value)
    if not original and not isclass(value):
        return False

    try:
        if original:
            return original and issubclass(original, _type)
        return issubclass(value, _type)
    except TypeError:
        return False



from .createapp import create_app as create_app  # noqa
from .createdeployment import create_deployment as create_deployment  # noqa
from .createproject import create_project as create_project  # noqa
from .list import list as list  # noqa
from .run import run as run  # noqa
from .runserver import runserver as runserver  # noqa
from .shell import shell as shell  # noqa
from .show_urls import show_urls as show_urls  # noqa



from __future__ import annotations

import json
from typing import Any, AsyncIterator, Iterable, cast

from lilya._internal._connection import Connection
from lilya.enums import Event, MessageMode, ScopeType, WebSocketState
from lilya.exceptions import WebSocketRuntimeError
from lilya.types import Message, Receive, Scope, Send


class WebSocketDisconnect(Exception):
    def __init__(self, code: int = 1000, reason: str | None = None) -> None:
        self.code = code
        self.reason = reason or ""


class WebsocketMixin(Connection):
    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope)
        assert scope["type"] == ScopeType.WEBSOCKET
        self._receive = receive
        self._send = send
        self._accepted = False
        self.client_state = WebSocketState.CONNECTING
        self.application_state = WebSocketState.CONNECTING

    def raise_for_disconnect(self, message: Message) -> None:
        if message["type"] == Event.WEBSOCKET_DISCONNECT:
            raise WebSocketDisconnect(message["code"], message.get("reason"))

    async def raise_for_connection_state(self) -> None:
        if self.application_state != WebSocketState.CONNECTED:
            raise WebSocketRuntimeError('WebSocket is not connected. Need to call "accept" first.')


class WebSocket(WebsocketMixin):
    async def receive(self) -> Message:
        """
        Receive ASGI websocket messages, ensuring valid state transitions.
        """
        if self.client_state not in {WebSocketState.CONNECTING, WebSocketState.CONNECTED}:
            raise WebSocketRuntimeError(
                'Cannot call "receive" once a disconnect message has been received.'
            )

        message = await self._receive()
        message_type = message["type"]

        if self.client_state == WebSocketState.CONNECTING:
            if message_type != Event.WEBSOCKET_CONNECT:
                raise WebSocketRuntimeError(
                    'Expected ASGI message "websocket.connect", ' f"but got {message_type!r}"
                )
            self.client_state = WebSocketState.CONNECTED
        elif self.client_state == WebSocketState.CONNECTED:
            if message_type not in {Event.WEBSOCKET_RECEIVE, Event.WEBSOCKET_DISCONNECT}:
                raise WebSocketRuntimeError(
                    "Expected ASGI message websocket.receive or "
                    f'"websocket.disconnect", but got {message_type!r}'
                )

            self.client_state = (
                WebSocketState.DISCONNECTED
                if message_type == Event.WEBSOCKET_DISCONNECT
                else self.client_state
            )
        return message

    async def send(self, message: Message) -> None:
        """
        Send ASGI websocket messages, ensuring valid state transitions.
        """
        if self.application_state not in {WebSocketState.CONNECTING, WebSocketState.CONNECTED}:
            raise WebSocketRuntimeError('Cannot call "send" once a close message has been sent.')

        message_type = message["type"]

        if self.application_state == WebSocketState.CONNECTING:
            if message_type not in {Event.WEBSOCKET_ACCEPT, Event.WEBSOCKET_CLOSE}:
                raise WebSocketRuntimeError(
                    'Expected ASGI message "websocket.accept" or '
                    f'"websocket.close", but got {message_type!r}'
                )
            self.application_state = (
                WebSocketState.DISCONNECTED
                if message_type == Event.WEBSOCKET_CLOSE
                else WebSocketState.CONNECTED
            )
        elif self.application_state == WebSocketState.CONNECTED:
            if message_type not in {Event.WEBSOCKET_SEND, Event.WEBSOCKET_CLOSE}:
                raise WebSocketRuntimeError(
                    'Expected ASGI message "websocket.send" or "websocket.close", '
                    f"but got {message_type!r}"
                )

            self.application_state = (
                WebSocketState.DISCONNECTED
                if message_type == Event.WEBSOCKET_CLOSE
                else self.application_state
            )
        await self._send(message)

    async def _connect(self) -> None:
        if self.client_state == WebSocketState.CONNECTING:
            message = await self._receive()
            assert message["type"] == Event.WEBSOCKET_CONNECT
        self.client_state = WebSocketState.CONNECTED

    async def accept(
        self,
        subprotocol: str | None = None,
        headers: Iterable[tuple[bytes, bytes]] | None = None,
    ) -> None:
        headers = headers or []

        await self._connect()
        message = {
            "type": "websocket.accept",
            "headers": headers,
            "subprotocol": subprotocol,
        }
        await self.send(message)

    async def receive_text(self) -> str:
        await self.raise_for_connection_state()

        message = await self.receive()
        self.raise_for_disconnect(message)
        return cast(str, message["text"])

    async def receive_bytes(self) -> bytes:
        await self.raise_for_connection_state()

        message = await self.receive()
        self.raise_for_disconnect(message)
        return cast(bytes, message["bytes"])

    async def receive_json(self, mode: str = "text") -> Any:
        if mode not in {MessageMode.TEXT, MessageMode.BINARY}:
            raise WebSocketRuntimeError('The "mode" argument should be "text" or "binary".')

        await self.raise_for_connection_state()

        message = await self.receive()
        self.raise_for_disconnect(message)

        if mode == "text":
            text = message["text"]
        else:
            text = message["bytes"].decode("utf-8")
        return json.loads(text)

    async def iter_text(self) -> AsyncIterator[str]:
        try:
            while True:
                yield await self.receive_text()
        except WebSocketDisconnect:
            ...  # pragma: no cover

    async def iter_bytes(self) -> AsyncIterator[bytes]:
        try:
            while True:
                yield await self.receive_bytes()
        except WebSocketDisconnect:
            ...  # pragma: no cover

    async def iter_json(self) -> AsyncIterator[Any]:
        try:
            while True:
                yield await self.receive_json()
        except WebSocketDisconnect:
            ...  # pragma: no cover

    async def send_text(self, data: str) -> None:
        await self.send({"type": Event.WEBSOCKET_SEND, "text": data})

    async def send_bytes(self, data: bytes) -> None:
        await self.send({"type": Event.WEBSOCKET_SEND, "bytes": data})

    async def send_json(self, data: Any, mode: str = "text") -> None:
        if mode not in {MessageMode.TEXT, MessageMode.BINARY}:
            raise WebSocketRuntimeError('The "mode" argument should be "text" or "binary".')

        text = json.dumps(data, separators=(",", ":"), ensure_ascii=False)

        (
            await self.send({"type": Event.WEBSOCKET_SEND, "text": text})
            if mode == MessageMode.TEXT
            else await self.send({"type": Event.WEBSOCKET_SEND, "bytes": text.encode("utf-8")})
        )

    async def close(self, code: int = 1000, reason: str | None = None) -> None:
        await self.send({"type": Event.WEBSOCKET_CLOSE, "code": code, "reason": reason or ""})


class WebSocketClose:
    def __init__(self, code: int = 1000, reason: str | None = None) -> None:
        self.code = code
        self.reason = reason or ""

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await send({"type": Event.WEBSOCKET_CLOSE, "code": self.code, "reason": self.reason})



from __future__ import annotations

from typing import TYPE_CHECKING

from lilya.conf import reload_settings, settings
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.types import ASGIApp, Receive, Scope, Send

if TYPE_CHECKING:
    from lilya.apps import Lilya


class ApplicationSettingsMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp):
        super().__init__(app)
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        app: Lilya = scope["app"]

        if getattr(app, "settings_module", None) is not None:
            settings.configure(app.settings)
        else:
            app_settings = reload_settings()
            settings.configure(app_settings())
        await self.app(scope, receive, send)



from __future__ import annotations

import os
import sys
from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar

import click

from lilya._internal._events import generate_lifespan_events
from lilya.cli.constants import APP_PARAMETER, LILYA_DISCOVER_APP
from lilya.cli.env import DirectiveEnv
from lilya.cli.terminal.print import Print
from lilya.cli.utils import fetch_directive
from lilya.compat import run_sync
from lilya.types import Lifespan

if TYPE_CHECKING:
    from lilya.apps import ChildLilya, Lilya

T = TypeVar("T")

printer = Print()


class Position(int, Enum):
    DEFAULT = 5
    BACK = 3


@click.option(
    "--directive",
    "directive",
    required=True,
    help=("The name of the file of the custom directive to run."),
)
@click.argument("directive_args", nargs=-1, type=click.UNPROCESSED)
@click.command(
    name="run",
    context_settings={
        "ignore_unknown_options": True,
    },
)
def run(env: DirectiveEnv, directive: str, directive_args: Any) -> None:
    """
    Runs every single custom directive in the system.

    How to run: `lilya --app <APP-LOCATION> run -n <DIRECTIVE NAME> <ARGS>`.

        Example: `lilya --app myapp:app run -n createsuperuser`
    """
    name = directive
    if name is not None and getattr(env, "app", None) is None:
        error = (
            "You cannot specify a custom directive without specifying the --app or setting "
            "LILYA_DEFAULT_APP environment variable."
        )
        printer.write_error(error)
        sys.exit(1)

    # Loads the directive object
    directive = fetch_directive(name, env.command_path, True)
    if not directive:
        printer.write_error(f"Unknown directive: {name!r}")
        sys.exit(1)

    # Execute the directive
    # The arguments for the directives start at the position 6
    position = get_position()
    program_name = " ".join(value for value in sys.argv[:position])

    ## Check if application is up and execute any event
    # Shutting down after
    lifespan = generate_lifespan_events(
        env.app.router.on_startup, env.app.router.on_shutdown, env.app.router.lifespan_context  # type: ignore
    )
    run_sync(execute_lifespan(env.app, lifespan, directive, program_name, position))


def get_position() -> int:
    """
    Gets the position of the arguments to read and pass them
    onto the directive.
    """
    if os.getenv(LILYA_DISCOVER_APP) is None and APP_PARAMETER in sys.argv:
        return Position.DEFAULT
    elif os.getenv(LILYA_DISCOVER_APP) is not None and APP_PARAMETER in sys.argv:
        return Position.DEFAULT
    return Position.BACK


async def execute_lifespan(
    app: Lilya | ChildLilya | None,
    lifespan: Lifespan,
    directive: Any,
    program_name: str,
    position: int,
) -> None:
    """
    Executes the lifespan events and the directive.
    """
    async with lifespan(app):
        await directive.execute_from_command(sys.argv[:], program_name, position)



from __future__ import annotations

from abc import ABC
from copy import copy
from dataclasses import asdict, dataclass
from functools import lru_cache
from http.cookies import SimpleCookie
from typing import (
    Any,
    BinaryIO,
    Generator,
    Generic,
    Iterable,
    Literal,
    Mapping,
    Sequence,
    TypeVar,
    cast,
)
from urllib.parse import SplitResult, parse_qsl, urlencode, urlsplit, urlunsplit

from anyio.to_thread import run_sync
from multidict import CIMultiDict, MultiDict as BaseMultiDict, MultiDictProxy, MultiMapping

from lilya.enums import DefaultPort, HTTPType, ScopeType, WebsocketType
from lilya.types import Scope

T = TypeVar("T")


class MultiMixin(Generic[T], MultiMapping[T], ABC):
    """Mixin providing common methods for multi dicts"""

    def dump(self) -> dict[str, Any]:
        """
        Returns the dictionary without duplicates.
        """
        return dict(self.items())

    def dump_list(self) -> list[Any]:
        """
        Returns the dictionary without duplicates.
        """
        return list(self.dump().items())

    def dict(self) -> dict[str, list[Any]]:
        """Return the multi-dict as a dict of lists."""
        return {k: self.getall(k) for k in set(self.keys())}

    def multi_items(self) -> Generator[tuple[str, T], None, None]:
        """Get all keys and values, including duplicates."""
        for key in set(self):
            for value in self.getall(key):
                yield key, value

    def get_multi_items(self) -> list[Any]:
        """
        Returns a list of values from the multi items
        """
        return list(self.multi_items())


class MultiDict(BaseMultiDict, MultiMixin[T], Generic[T]):
    def __init__(
        self,
        args: MultiMapping | Mapping[str, T] | Iterable[tuple[str, T]] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(args or {})

    def to_immutable(self) -> ImmutableMultiDict[T]:
        """Create an immutable dictionary view."""
        return ImmutableMultiDict[T](self)

    def getlist(self, key: Any) -> list[Any]:
        return [item_value for item_key, item_value in list(self.multi_items()) if item_key == key]

    def poplist(self, key: Any) -> list[Any]:
        values = [v for k, v in list(self.multi_items()) if k == key]
        self.pop(key)
        return values

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return sorted(self.multi_items()) == sorted(other.multi_items())

    def __repr__(self) -> str:
        """
        Visual representation of the object sorted by keys.
        """
        items = sorted(self.items())
        return f"{self.__class__.__name__}({items!r})"


class ImmutableMultiDict(MultiDictProxy[T], MultiMixin[T], Generic[T]):
    def __init__(
        self,
        args: MultiMapping | Mapping[str, Any] | Iterable[tuple[str, Any]] | Any | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(BaseMultiDict(args or {}))

    def mutablecopy(self) -> MultiDict[T]:
        """Create a mutable copy as a.

        Returns:
            A mutable MultiDict.
        """
        return MultiDict(list(self.multi_items()))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return sorted(self.multi_items()) == sorted(other.multi_items())

    def __repr__(self) -> str:
        """
        Visual representation of the object sorted by keys.
        """
        items = sorted(self.items())
        return f"{self.__class__.__name__}({items!r})"


class FormMultiDict(ImmutableMultiDict[Any]):
    """MultiDict for form data."""

    async def close(self) -> None:
        """Close all files in the FormMultiDict."""
        for _, value in self.multi_items():
            if isinstance(value, DataUpload):
                await value.close()


class Header(MultiDict, CIMultiDict):  # type: ignore
    """Container used for both request and response headers.
    It is a subclass of  [CIMultiDict](https://multidict.readthedocs.io/en/stable/multidict.html#cimultidictproxy)

    Please checkout [the MultiDict documentation](https://multidict.readthedocs.io/en/stable/multidict.html#multidict) for more details.
    """

    def __init__(
        self,
        *args: MultiMapping
        | Mapping[str, Any]
        | Iterable[tuple[str, Any]]
        | list[tuple[bytes, bytes]]
        | None,
    ) -> None:
        assert len(args) < 2, "Too many arguments."
        value = args[0] if args else []

        assert isinstance(
            value, (dict, list)
        ), "The headers must be in the format of a list of tuples or dicttionary."

        headers: list[tuple[str, Any]] = self.parse_headers(value)
        super().__init__(headers)

    def parse_headers(self, value: Any) -> list[tuple[str, Any]]:
        """
        Parses the headers and validates if its bytes or str.
        """
        headers: list[tuple[str, Any]] = []

        if isinstance(value, dict):
            for k, v in value.items():
                key = k.decode("latin-1") if isinstance(k, bytes) else k
                value = v.decode("latin-1") if isinstance(v, bytes) else v
                headers.append((key, value))
        if isinstance(value, list):
            for k, v in value:
                key = k.decode("latin-1") if isinstance(k, bytes) else k
                value = v.decode("latin-1") if isinstance(v, bytes) else v
                headers.append((key, value))

        return headers

    def __getattr__(self, key: str) -> str:
        if key.startswith("_"):
            return cast(str, self.__getattribute__(key))
        key = key.rstrip("_").replace("_", "-")
        return ",".join(self.getall(key, default=[]))

    def get_all(self, key: str) -> list[Any]:
        """Convenience method mapped to getall()."""
        return self.getall(key, default=[])

    @classmethod
    def from_scope(cls, scope: Any) -> Header:
        """
        Builds the headers from the scope.
        """
        return cls(scope["headers"])

    def add_vary_header(self, vary: str) -> None:
        existing = self.get("vary")
        if existing is not None:
            vary = ", ".join([existing, vary])
        self["vary"] = vary

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        as_dict = dict(self.items())
        if len(as_dict) == len(self):
            return f"{class_name}({as_dict!r})"
        return class_name


class State:
    """
    An object that can be used to store arbitrary state.

    Used for `request.state` and `app.state`.
    """

    _state: dict[str, Any]

    def __init__(self, state: dict[str, Any] | None = None):
        if state is None:
            state = {}
        super().__setattr__("_state", state)

    def __setattr__(self, key: Any, value: Any) -> None:
        self._state[key] = value

    def __getattr__(self, key: Any) -> Any:
        try:
            return self._state[key]
        except KeyError:
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, key)) from None

    def __delattr__(self, key: Any) -> None:
        del self._state[key]

    def __copy__(self) -> State:
        return self.__class__(copy(self._state))

    def __len__(self) -> int:
        return len(self._state)

    def __getitem__(self, key: str) -> Any:
        return self._state[key]

    def copy(self) -> State:
        return copy(self)


class URL:
    """
    Represents a URL and provides methods for manipulating it.
    """

    url: str
    scheme: str | None = None
    netloc: str | None = None
    path: str | None = None
    fragment: str | None = None
    query: str | None = None
    username: str | None = None
    password: str | None = None
    port: int | None = None
    hostname: str | None = None

    def __new__(cls, url: str | SplitResult | None = None, **kwargs: Any) -> URL:
        """
        Overrides the new base class to create a new structure for the URL.

        Args:
            url (str | SplitResult | None): The URL to create.

        Returns:
            URL: The created URL instance.
        """
        instance = cls.__create__(url) if url is not None else cls.__create_from_kwargs__(**kwargs)
        return instance

    @classmethod
    @lru_cache
    def __create_from_kwargs__(cls, **kwargs: Any) -> URL:
        """
        Creates a new URL instance from keyword arguments.

        Args:
            **kwargs: Keyword arguments representing URL components.

        Returns:
            URL: The created URL instance.
        """
        instance: URL = super().__new__(cls)

        for k, _ in instance.__annotations__.items():
            if k in kwargs:
                setattr(instance, k, kwargs[k])
        instance.url = instance.base_url
        return instance

    @classmethod
    @lru_cache
    def __create__(cls, url: str | SplitResult | None = None) -> URL:
        """
        Creates a new URL instance.

        Args:
            url (str | SplitResult | None): The URL to create.

        Returns:
            URL: The created URL instance.
        """
        instance: URL = super().__new__(cls)

        assert isinstance(
            url, (str, SplitResult)
        ), "The url must be a string or a SplitResult instance."

        if isinstance(url, str):
            result: SplitResult = urlsplit(url)
            instance.parsed_url = result
        else:
            instance.parsed_url = url

        instance.url = instance.base_url if isinstance(url, SplitResult) else url
        instance.scheme = instance.parsed_url.scheme
        instance.netloc = instance.parsed_url.netloc
        instance.path = instance.parsed_url.path
        instance.fragment = instance.parsed_url.fragment
        instance.query = instance.parsed_url.query
        instance.username = instance.parsed_url.username
        instance.password = instance.parsed_url.password
        instance.port = instance.parsed_url.port
        instance.hostname = instance.parsed_url.hostname
        return instance

    @classmethod
    def build_from_scope(cls, scope: Scope) -> URL:
        """
        Builds the URL from the Scope.

        Args:
            scope (Scope): The scope object.

        Returns:
            URL: The URL built from the scope.
        """
        scheme = scope.get("scheme", HTTPType.HTTP.value)
        server = scope.get("server")
        path = scope["path"]
        query_string = scope.get("query_string", b"")

        host_header = None
        for key, value in scope["headers"]:
            if key == b"host":
                host_header = value.decode("latin-1")
                break

        if host_header is not None:
            url = f"{scheme}://{host_header}{path}"
        elif server is None:
            url = path
        else:
            host, port = server
            default_port = DefaultPort.to_dict()[scheme]
            if port == default_port:
                url = f"{scheme}://{host}{path}"
            else:
                url = f"{scheme}://{host}:{port}{path}"

        if query_string:
            url += "?" + query_string.decode()
        return cls(url)

    @property
    def parsed_url(self) -> SplitResult:
        """
        Property representing the parsed URL.

        Returns:
            SplitResult: The parsed URL.
        """
        return cast(SplitResult, getattr(self, "_parsed_url", None))

    @parsed_url.setter
    def parsed_url(self, value: Any) -> None:
        """
        Setter for the parsed URL.

        Args:
            value: The value to set as the parsed URL.
        """
        self._parsed_url = value

    @property
    def is_secure(self) -> bool:
        """
        Property indicating whether the URL uses a secure scheme.

        Returns:
            bool: True if the URL is secure, False otherwise.
        """
        return self.scheme in ("https", "wss")

    @property
    def base_url(self) -> str:
        """
        Property representing the base URL.

        Returns:
            str: The base URL.
        """
        return str(
            urlunsplit(
                SplitResult(
                    scheme=self.scheme,
                    netloc=self.netloc,
                    path=self.path,
                    fragment=self.fragment,
                    query=self.query,
                )
            )
        )

    def _build_netloc(
        self,
        hostname: Any | None = None,
        port: Any | None = None,
        username: Any | None = None,
        password: Any | None = None,
    ) -> Any:
        """
        Builds the netloc part of the URL.

        Args:
            hostname (Any | None): The hostname.
            port (Any | None): The port.
            username (Any | None): The username.
            password (Any | None): The password.

        Returns:
            Any: The built netloc.
        """
        if hostname is None:
            netloc = self.netloc
            _, _, hostname = netloc.rpartition("@")

            if hostname[-1] != "]":
                hostname = hostname.rsplit(":", 1)[0]

        netloc = hostname
        if port is not None:
            netloc += f":{port}"
        if username is not None:
            userpass = username
            if password is not None:
                userpass += f":{password}"
            netloc = f"{userpass}@{netloc}"
        return netloc

    def replace(self, **kwargs: Any) -> URL:
        """
        Creates a new URL with replaced components.

        Args:
            **kwargs: Keyword arguments representing URL components.

        Returns:
            URL: The new URL instance with replaced components.
        """
        if (
            "username" in kwargs
            or "password" in kwargs
            or "hostname" in kwargs
            or "port" in kwargs
        ):
            hostname = kwargs.pop("hostname", None)
            port = kwargs.pop("port", self.port)
            username = kwargs.pop("username", self.username)
            password = kwargs.pop("password", self.password)
            kwargs["netloc"] = self._build_netloc(hostname, port, username, password)

        components = self.parsed_url._replace(**kwargs)
        return self.__class__(components.geturl())

    def include_query_params(self, **kwargs: Any) -> URL:
        """
        Adds query parameters to the URL.

        Args:
            **kwargs: Query parameters.

        Returns:
            URL: The new URL instance with added query parameters.
        """
        params = MultiDict(parse_qsl(self.query, keep_blank_values=True))
        params.update({str(key): str(value) for key, value in kwargs.items()})
        query = urlencode(sorted(params.multi_items()))
        return self.replace(query=query)

    def remove_query_params(self, keys: str | Sequence[str]) -> URL:
        """
        Removes query parameters from the URL.

        Args:
            keys (str | Sequence[str]): Query parameters to remove.

        Returns:
            URL: The new URL instance with removed query parameters.
        """
        if isinstance(keys, str):
            keys = [keys]
        params = MultiDict(parse_qsl(self.query, keep_blank_values=True))
        for key in keys:
            params.pop(key, None)
        query = urlencode(sorted(params.multi_items()))
        return self.replace(query=query)

    def replace_query_params(self, **kwargs: Any) -> URL:
        """
        Replaces query parameters in the URL.

        Args:
            **kwargs: Query parameters to replace.

        Returns:
            URL: The new URL instance with replaced query parameters.
        """
        values = [(str(key), str(value)) for key, value in kwargs.items()]
        query = urlencode(values)
        return self.replace(query=query)

    def __eq__(self, other: Any) -> bool:
        """
        Checks if two URLs are equal.

        Args:
            other: The other object to compare.

        Returns:
            bool: True if the URLs are equal, False otherwise.
        """
        return str(self) == str(other)

    def __str__(self) -> str:
        """
        Returns the string representation of the URL.

        Returns:
            str: The string representation of the URL.
        """
        return self.url

    def __repr__(self) -> str:
        """
        Returns the string representation of the URL for debugging purposes.

        Returns:
            str: The string representation of the URL.
        """
        url = str(self)
        if self.password:
            url = str(self.replace(password="***********"))
        return f"{self.__class__.__name__}({repr(url)})"


class URLPath(str):
    def __new__(cls, path: str, protocol: str = "", host: str = "") -> URLPath:
        assert protocol in (ScopeType.HTTP.value, ScopeType.WEBSOCKET.value, "")
        return str.__new__(cls, path)

    def __init__(self, path: str, protocol: str = "", host: str = "") -> None:
        self.protocol = protocol
        self.host = host
        self.path = path

    def make_absolute_url(self, base_url: str | URL) -> URL:
        if isinstance(base_url, str):
            base_url = URL(base_url)
        if self.protocol:
            scheme = {
                ScopeType.HTTP.value: {True: HTTPType.HTTPS.value, False: HTTPType.HTTP.value},
                ScopeType.WEBSOCKET.value: {
                    True: WebsocketType.WSS.value,
                    False: WebsocketType.WS.value,
                },
            }[self.protocol][base_url.is_secure]
        else:
            scheme = base_url.scheme

        netloc = self.host or base_url.netloc
        path = base_url.path.rstrip("/") + str(self)
        return URL(scheme=scheme, netloc=netloc, path=path)


@dataclass
class Secret:
    """
    Holds the information about a string secret.
    This will make sure no secrets are leaked
    in stack traces.
    """

    value: str = None

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('***********')"

    def __bool__(self) -> bool:
        return bool(self.value)

    def __len__(self) -> int:
        return len(self.value)


class QueryParam(ImmutableMultiDict[Any]):
    """
    An immutable multidict.
    """

    def __init__(
        self,
        *args: MultiMapping | Mapping[str, Any] | Iterable[tuple[str, Any]] | None,
    ) -> None:
        assert len(args) < 2, "Too many arguments."
        value = args[0] if args else []

        if isinstance(value, str):  # type: ignore
            super().__init__(parse_qsl(value, keep_blank_values=True))  # type: ignore
        elif isinstance(value, bytes):  # type: ignore
            super().__init__(parse_qsl(value.decode("latin-1"), keep_blank_values=True))  # type: ignore
        else:
            super().__init__(*args)

    def __str__(self) -> str:
        return urlencode(sorted(self.multi_items()))

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        query_string = str(self)
        return f"{class_name}({query_string!r})"


class DataUpload:
    __slots__ = ("file", "size", "filename", "headers")

    def __init__(
        self,
        *,
        file: BinaryIO,
        size: int | None = None,
        filename: str | None = None,
        headers: Header | None = None,
    ) -> None:
        self.filename = filename
        self.size = size
        self.headers = headers or Header()
        self.file = file

    @property
    def content_type(self) -> str | None:
        return self.headers.get("content-type", None)

    @property
    def in_memory(self) -> bool:
        return getattr(self.file, "_rolled", False)

    async def write(self, data: bytes) -> int:
        if self.size is not None:
            self.size += len(data)

        if self.in_memory:
            return await run_sync(self.file.write, data)
        return self.file.write(data)

    async def read(self, size: int = -1) -> bytes:
        if self.in_memory:
            return self.file.read(size)
        return await run_sync(self.file.read, size)

    async def seek(self, offset: int) -> int:
        if self.in_memory:
            return await run_sync(self.file.seek, offset)
        return self.file.seek(offset)

    async def close(self) -> int:
        if self.in_memory:
            return await run_sync(self.file.close)
        return self.file.close()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"filename={self.filename!r}, "
            f"size={self.size!r}, "
            f"headers={self.headers!r})"
        )


@dataclass
class Cookie:
    key: str
    value: str | None = None
    max_age: int | None = None
    expires: int | None = None
    path: str = "/"
    domain: str | None = None
    secure: bool | None = None
    httponly: bool | None = None
    samesite: Literal["lax", "strict", "none"] = "lax"
    description: str | None = None

    def to_header(self, **kwargs: Any) -> str:
        simple_cookie: SimpleCookie = SimpleCookie()
        simple_cookie[self.key] = self.value or ""
        if self.max_age:
            simple_cookie[self.key]["max-age"] = self.max_age
        cookie_dict = asdict(self)
        for key in ["expires", "path", "domain", "secure", "httponly", "samesite"]:
            if cookie_dict[key] is not None:
                simple_cookie[self.key][key] = cookie_dict[key]
        return simple_cookie.output(**kwargs).strip()


class FormData(ImmutableMultiDict[Any]):
    """
    An immutable multidict, containing both file uploads and text input.
    """

    def __init__(
        self,
        *args: FormData | Mapping[str, str | DataUpload] | list[tuple[str, str | DataUpload]],
        **kwargs: str | DataUpload,
    ) -> None:
        super().__init__(*args, **kwargs)

    async def close(self) -> None:
        for _, value in self.multi_items():
            if isinstance(value, DataUpload):
                await value.close()



from __future__ import annotations

import argparse
import os
import sys
from abc import ABC, abstractmethod
from typing import Any

import lilya
from lilya.cli.exceptions import DirectiveError
from lilya.cli.parsers import DirectiveParser
from lilya.cli.terminal.print import Print
from lilya.compat import is_async_callable

printer = Print()


class BaseDirective(ABC):
    """The base class from which all directrives derive"""

    help: str = ""

    def get_version(self) -> str:
        """
        Returns the current version of Lilya.
        """
        return lilya.__version__

    def add_arguments(self, parser: type[argparse.ArgumentParser]) -> Any:  # noqa
        """
        Entrypoint for directives and custom arguments
        """
        ...

    def create_parser(self, name: str, subdirective: str, **kwargs: Any) -> DirectiveParser:
        parser = DirectiveParser(
            prog=f"{os.path.basename(name)} {subdirective}",
            description=self.help,
            **kwargs,
        )
        self.add_arguments(parser)  # type: ignore
        return parser

    async def execute_from_command(self, argv: Any, program_name: str, position: int = 5) -> None:
        """
        Executes dynamically the directive from the command line and passing the parameters
        """
        parser = self.create_parser(program_name, argv[position])

        options = parser.parse_args(argv[position + 1 :])
        cmd_options = vars(options)

        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop("args", ())
        try:
            await self.run(*args, **cmd_options)
        except DirectiveError as e:
            printer.write_error(f"{e.__class__.__name__}: {e}")
            sys.exit(e.returncode)

    async def run(self, *args: Any, **options: Any) -> Any:
        """
        Executes the handle()
        """
        if not is_async_callable(self.handle):
            output = self.handle(*args, **options)
        else:
            output = await self.handle(*args, **options)
        if output:
            printer.write_info(output)
        return output

    @abstractmethod
    def handle(self, *args: Any, **options: Any) -> Any:
        """The logic of the directive. Subclasses must implement this method"""
        raise NotImplementedError("subclasses of BaseDirective must provide a handle() method.")



"""
Copyright © 2018, [Encode OSS Ltd](https://www.encode.io/).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This file contains some adaptations used for Lilya purposes.
The nature of the test client is exactly the same as the one
provided by Starlette, therefore, Lilya reuses it.
"""

from __future__ import annotations

import contextlib
import inspect
import io
import json
import math
import queue
import sys
import typing
import warnings
from concurrent.futures import Future
from functools import cached_property
from types import GeneratorType
from urllib.parse import unquote, urljoin

import anyio
import anyio.from_thread
from anyio.abc import ObjectReceiveStream, ObjectSendStream
from anyio.streams.stapled import StapledObjectStream

from lilya.compat import is_async_callable
from lilya.types import ASGIApp, Message, Receive, Scope, Send
from lilya.websockets import WebSocketDisconnect

try:
    import httpx
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        "The lilya.testclient module requires the httpx package to be installed.\n"
        "You can install this with:\n"
        "    $ pip install httpx\n"
    ) from None


if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import TypeGuard
else:  # pragma: no cover
    from typing_extensions import TypeGuard


_PortalFactoryType = typing.Callable[[], typing.ContextManager[anyio.abc.BlockingPortal]]

ASGIInstance = typing.Callable[[Receive, Send], typing.Awaitable[None]]
ASGI2App = typing.Callable[[Scope], ASGIInstance]
ASGI3App = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]


_RequestData = typing.Mapping[str, typing.Union[str, typing.Iterable[str]]]


def _is_asgi3(app: ASGI2App | ASGI3App) -> TypeGuard[ASGI3App]:
    if inspect.isclass(app):
        return hasattr(app, "__await__")
    return is_async_callable(app)


class _WrapASGI2:
    """
    Provide an ASGI3 interface onto an ASGI2 app.
    """

    def __init__(self, app: ASGI2App) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        instance = self.app(scope)
        await instance(receive, send)


class _AsyncBackend(typing.TypedDict):
    backend: str
    backend_options: dict[str, typing.Any]


class _Upgrade(Exception):
    def __init__(self, session: WebSocketTestSession) -> None:
        self.session = session


class WebSocketDenialResponse(
    httpx.Response,
    WebSocketDisconnect,
):
    """
    A special case of `WebSocketDisconnect`, raised in the `TestClient` if the
    `WebSocket` is closed before being accepted with a `send_denial_response()`.
    """

    ...


class WebSocketTestSession:
    def __init__(
        self,
        app: ASGI3App,
        scope: Scope,
        portal_factory: _PortalFactoryType,
    ) -> None:
        self.app = app
        self.scope = scope
        self.accepted_subprotocol = None
        self.portal_factory = portal_factory
        self._receive_queue: queue.Queue[Message] = queue.Queue()
        self._send_queue: queue.Queue[Message | BaseException] = queue.Queue()
        self.extra_headers = None

    def __enter__(self) -> WebSocketTestSession:
        self.exit_stack = contextlib.ExitStack()
        self.portal = self.exit_stack.enter_context(self.portal_factory())

        try:
            _: Future[None] = self.portal.start_task_soon(self._run)
            self.send({"type": "websocket.connect"})
            message = self.receive()
            self._raise_on_close(message)
        except Exception:
            self.exit_stack.close()
            raise
        self.accepted_subprotocol = message.get("subprotocol", None)
        self.extra_headers = message.get("headers", None)
        return self

    @cached_property
    def should_close(self) -> anyio.Event:
        return anyio.Event()

    async def _notify_close(self) -> None:
        self.should_close.set()

    def __exit__(self, *args: typing.Any) -> None:
        try:
            self.close(1000)
        finally:
            self.portal.start_task_soon(self._notify_close)
            self.exit_stack.close()
        while not self._send_queue.empty():
            message = self._send_queue.get()
            if isinstance(message, BaseException):
                raise message

    async def _run(self) -> None:
        """
        The sub-thread in which the websocket session runs.
        """

        async def run_app(tg: anyio.abc.TaskGroup) -> None:
            try:
                await self.app(self.scope, self._asgi_receive, self._asgi_send)
            except anyio.get_cancelled_exc_class():
                ...
            except BaseException as exc:
                self._send_queue.put(exc)
                raise
            finally:
                tg.cancel_scope.cancel()

        async with anyio.create_task_group() as tg:
            tg.start_soon(run_app, tg)
            await self.should_close.wait()
            tg.cancel_scope.cancel()

    async def _asgi_receive(self) -> Message:
        while self._receive_queue.empty():
            await anyio.sleep(0)
        return self._receive_queue.get()

    async def _asgi_send(self, message: Message) -> None:
        self._send_queue.put(message)

    def _raise_on_close(self, message: Message) -> None:
        if message["type"] == "websocket.close":
            raise WebSocketDisconnect(
                code=message.get("code", 1000), reason=message.get("reason", "")
            )
        elif message["type"] == "websocket.http.response.start":
            status_code: int = message["status"]
            headers: list[tuple[bytes, bytes]] = message["headers"]
            body: list[bytes] = []
            while True:
                message = self.receive()
                assert message["type"] == "websocket.http.response.body"
                body.append(message["body"])
                if not message.get("more_body", False):
                    break
            raise WebSocketDenialResponse(
                status_code=status_code,
                headers=headers,
                content=b"".join(body),
            )

    def send(self, message: Message) -> None:
        self._receive_queue.put(message)

    def send_text(self, data: str) -> None:
        self.send({"type": "websocket.receive", "text": data})

    def send_bytes(self, data: bytes) -> None:
        self.send({"type": "websocket.receive", "bytes": data})

    def send_json(self, data: typing.Any, mode: typing.Literal["text", "binary"] = "text") -> None:
        text = json.dumps(data, separators=(",", ":"), ensure_ascii=False)
        if mode == "text":
            self.send({"type": "websocket.receive", "text": text})
        else:
            self.send({"type": "websocket.receive", "bytes": text.encode("utf-8")})

    def close(self, code: int = 1000, reason: str | None = None) -> None:
        self.send({"type": "websocket.disconnect", "code": code, "reason": reason})

    def receive(self) -> Message:
        message = self._send_queue.get()
        if isinstance(message, BaseException):
            raise message
        return message

    def receive_text(self) -> str:
        message = self.receive()
        self._raise_on_close(message)
        return typing.cast(str, message["text"])

    def receive_bytes(self) -> bytes:
        message = self.receive()
        self._raise_on_close(message)
        return typing.cast(bytes, message["bytes"])

    def receive_json(self, mode: typing.Literal["text", "binary"] = "text") -> typing.Any:
        message = self.receive()
        self._raise_on_close(message)
        if mode == "text":
            text = message["text"]
        else:
            text = message["bytes"].decode("utf-8")
        return json.loads(text)


class _TestClientTransport(httpx.BaseTransport):
    def __init__(
        self,
        app: ASGI3App,
        portal_factory: _PortalFactoryType,
        raise_server_exceptions: bool = True,
        root_path: str = "",
        *,
        app_state: dict[str, typing.Any],
    ) -> None:
        self.app = app
        self.raise_server_exceptions = raise_server_exceptions
        self.root_path = root_path
        self.portal_factory = portal_factory
        self.app_state = app_state

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        scheme = request.url.scheme
        netloc = request.url.netloc.decode(encoding="ascii")
        path = request.url.path
        raw_path = request.url.raw_path
        query = request.url.query.decode(encoding="ascii")

        default_port = {"http": 80, "ws": 80, "https": 443, "wss": 443}[scheme]

        if ":" in netloc:
            host, port_string = netloc.split(":", 1)
            port = int(port_string)
        else:
            host = netloc
            port = default_port

        # Include the 'host' header.
        if "host" in request.headers:
            headers: list[tuple[bytes, bytes]] = []
        elif port == default_port:  # pragma: no cover
            headers = [(b"host", host.encode())]
        else:  # pragma: no cover
            headers = [(b"host", (f"{host}:{port}").encode())]

        # Include other request headers.
        headers += [
            (key.lower().encode(), value.encode()) for key, value in request.headers.multi_items()
        ]

        scope: dict[str, typing.Any]

        if scheme in {"ws", "wss"}:
            subprotocol = request.headers.get("sec-websocket-protocol", None)
            if subprotocol is None:
                subprotocols: typing.Sequence[str] = []
            else:
                subprotocols = [value.strip() for value in subprotocol.split(",")]
            scope = {
                "type": "websocket",
                "path": unquote(path),
                "raw_path": raw_path,
                "root_path": self.root_path,
                "scheme": scheme,
                "query_string": query.encode(),
                "headers": headers,
                "client": None,
                "server": [host, port],
                "subprotocols": subprotocols,
                "state": self.app_state.copy(),
                "extensions": {"websocket.http.response": {}},
            }
            session = WebSocketTestSession(self.app, scope, self.portal_factory)
            raise _Upgrade(session)

        scope = {
            "type": "http",
            "http_version": "1.1",
            "method": request.method,
            "path": unquote(path),
            "raw_path": raw_path,
            "root_path": self.root_path,
            "scheme": scheme,
            "query_string": query.encode(),
            "headers": headers,
            "client": None,
            "server": [host, port],
            "extensions": {"http.response.debug": {}},
            "state": self.app_state.copy(),
        }

        request_complete = False
        response_started = False
        response_complete: anyio.Event
        raw_kwargs: dict[str, typing.Any] = {"stream": io.BytesIO()}
        template = None
        context = None

        async def receive() -> Message:
            nonlocal request_complete

            if request_complete:
                if not response_complete.is_set():
                    await response_complete.wait()
                return {"type": "http.disconnect"}

            body = request.read()
            if isinstance(body, str):  # type: ignore
                body_bytes: bytes = body.encode("utf-8")  # type: ignore
            elif body is None:
                body_bytes = b""  # pragma: no cover
            elif isinstance(body, GeneratorType):  # type: ignore
                try:  # type: ignore
                    chunk = body.send(None)
                    if isinstance(chunk, str):
                        chunk = chunk.encode("utf-8")
                    return {"type": "http.request", "body": chunk, "more_body": True}
                except StopIteration:  # pragma: no cover
                    request_complete = True
                    return {"type": "http.request", "body": b""}
            else:
                body_bytes = body

            request_complete = True
            return {"type": "http.request", "body": body_bytes}

        async def send(message: Message) -> None:
            nonlocal raw_kwargs, response_started, template, context

            if message["type"] == "http.response.start":
                assert not response_started, 'Received multiple "http.response.start" messages.'
                raw_kwargs["status_code"] = message["status"]
                raw_kwargs["headers"] = list(message.get("headers", []))
                response_started = True
            elif message["type"] == "http.response.body":
                assert (
                    response_started
                ), 'Received "http.response.body" without "http.response.start".'
                assert (
                    not response_complete.is_set()
                ), 'Received "http.response.body" after response completed.'
                body = message.get("body", b"")
                more_body = message.get("more_body", False)
                if request.method != "HEAD":
                    raw_kwargs["stream"].write(body)
                if not more_body:
                    raw_kwargs["stream"].seek(0)
                    response_complete.set()
            elif message["type"] == "http.response.debug":
                template = message["info"]["template"]
                context = message["info"]["context"]

        try:
            with self.portal_factory() as portal:
                response_complete = portal.call(anyio.Event)
                portal.call(self.app, scope, receive, send)
        except BaseException as exc:
            if self.raise_server_exceptions:
                raise exc

        if self.raise_server_exceptions:
            assert response_started, "TestClient did not receive any response."
        elif not response_started:
            raw_kwargs = {
                "status_code": 500,
                "headers": [],
                "stream": io.BytesIO(),
            }

        raw_kwargs["stream"] = httpx.ByteStream(raw_kwargs["stream"].read())

        response = httpx.Response(**raw_kwargs, request=request)
        if template is not None:
            response.template = template  # type: ignore[attr-defined]
            response.context = context  # type: ignore[attr-defined]
        return response


class TestClient(httpx.Client):
    __test__ = False
    task: Future[None]
    portal: anyio.abc.BlockingPortal | None = None

    def __init__(
        self,
        app: ASGIApp,
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
        backend: typing.Literal["asyncio", "trio"] = "asyncio",
        backend_options: dict[str, typing.Any] | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        headers: dict[str, str] | None = None,
        follow_redirects: bool = True,
    ) -> None:
        self.async_backend = _AsyncBackend(backend=backend, backend_options=backend_options or {})
        if _is_asgi3(app):
            asgi_app = app
        else:
            app = typing.cast(ASGI2App, app)  # type: ignore[assignment]
            asgi_app = _WrapASGI2(app)  # type: ignore[arg-type]
        self.app = asgi_app
        self.app_state: dict[str, typing.Any] = {}
        transport = _TestClientTransport(
            self.app,
            portal_factory=self._portal_factory,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            app_state=self.app_state,
        )
        if headers is None:
            headers = {}
        headers.setdefault("user-agent", "testclient")
        super().__init__(
            app=self.app,
            base_url=base_url,
            headers=headers,
            transport=transport,
            follow_redirects=follow_redirects,
            cookies=cookies,
        )

    @contextlib.contextmanager
    def _portal_factory(self) -> typing.Generator[anyio.abc.BlockingPortal, None, None]:
        if self.portal is not None:
            yield self.portal
        else:
            with anyio.from_thread.start_blocking_portal(**self.async_backend) as portal:
                yield portal

    def _choose_redirect_arg(
        self, follow_redirects: bool | None, allow_redirects: bool | None
    ) -> bool | httpx._client.UseClientDefault:
        redirect: bool | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT
        if allow_redirects is not None:
            message = (
                "The `allow_redirects` argument is deprecated. " "Use `follow_redirects` instead."
            )
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            redirect = allow_redirects
        if follow_redirects is not None:
            redirect = follow_redirects
        elif allow_redirects is not None and follow_redirects is not None:
            raise RuntimeError(  # pragma: no cover
                "Cannot use both `allow_redirects` and `follow_redirects`."
            )
        return redirect

    def request(  # type: ignore[override]
        self,
        method: str,
        url: httpx._types.URLTypes,
        *,
        content: httpx._types.RequestContent | None = None,
        data: _RequestData | None = None,
        files: httpx._types.RequestFiles | None = None,
        json: typing.Any = None,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        url = self._merge_url(url)
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def get(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().get(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def options(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().options(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def head(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().head(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def post(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        content: httpx._types.RequestContent | None = None,
        data: _RequestData | None = None,
        files: httpx._types.RequestFiles | None = None,
        json: typing.Any = None,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().post(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def put(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        content: httpx._types.RequestContent | None = None,
        data: _RequestData | None = None,
        files: httpx._types.RequestFiles | None = None,
        json: typing.Any = None,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().put(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def patch(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        content: httpx._types.RequestContent | None = None,
        data: _RequestData | None = None,
        files: httpx._types.RequestFiles | None = None,
        json: typing.Any = None,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().patch(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def delete(  # type: ignore[override]
        self,
        url: httpx._types.URLTypes,
        *,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        auth: (
            httpx._types.AuthTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        allow_redirects: bool | None = None,
        timeout: (
            httpx._types.TimeoutTypes | httpx._client.UseClientDefault
        ) = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, typing.Any] | None = None,
    ) -> httpx.Response:
        redirect = self._choose_redirect_arg(follow_redirects, allow_redirects)
        return super().delete(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def websocket_connect(
        self,
        url: str,
        subprotocols: typing.Sequence[str] | None = None,
        **kwargs: typing.Any,
    ) -> WebSocketTestSession:
        url = urljoin("ws://testserver", url)
        headers = kwargs.get("headers", {})
        headers.setdefault("connection", "upgrade")
        headers.setdefault("sec-websocket-key", "testserver==")
        headers.setdefault("sec-websocket-version", "13")
        if subprotocols is not None:
            headers.setdefault("sec-websocket-protocol", ", ".join(subprotocols))
        kwargs["headers"] = headers
        try:
            super().request("GET", url, **kwargs)
        except _Upgrade as exc:
            session = exc.session
        else:
            raise RuntimeError("Expected WebSocket upgrade")  # pragma: no cover

        return session

    def __enter__(self) -> TestClient:
        with contextlib.ExitStack() as stack:
            self.portal = portal = stack.enter_context(
                anyio.from_thread.start_blocking_portal(**self.async_backend)
            )

            @stack.callback
            def reset_portal() -> None:
                self.portal = None

            send1: ObjectSendStream[typing.MutableMapping[str, typing.Any] | None]
            receive1: ObjectReceiveStream[typing.MutableMapping[str, typing.Any] | None]
            send2: ObjectSendStream[typing.MutableMapping[str, typing.Any]]
            receive2: ObjectReceiveStream[typing.MutableMapping[str, typing.Any]]
            send1, receive1 = anyio.create_memory_object_stream(math.inf)
            send2, receive2 = anyio.create_memory_object_stream(math.inf)
            self.stream_send = StapledObjectStream(send1, receive1)
            self.stream_receive = StapledObjectStream(send2, receive2)
            self.task = portal.start_task_soon(self.lifespan)
            portal.call(self.wait_startup)

            @stack.callback
            def wait_shutdown() -> None:
                portal.call(self.wait_shutdown)

            self.exit_stack = stack.pop_all()

        return self

    def __exit__(self, *args: typing.Any) -> None:
        self.exit_stack.close()

    async def lifespan(self) -> None:
        scope = {"type": "lifespan", "state": self.app_state}
        try:
            await self.app(scope, self.stream_receive.receive, self.stream_send.send)
        finally:
            await self.stream_send.send(None)

    async def wait_startup(self) -> None:
        await self.stream_receive.send({"type": "lifespan.startup"})

        async def receive() -> typing.Any:
            message = await self.stream_send.receive()
            if message is None:
                self.task.result()
            return message

        message = await receive()
        assert message["type"] in (
            "lifespan.startup.complete",
            "lifespan.startup.failed",
        )
        if message["type"] == "lifespan.startup.failed":
            await receive()

    async def wait_shutdown(self) -> None:
        async def receive() -> typing.Any:
            message = await self.stream_send.receive()
            if message is None:
                self.task.result()
            return message

        async with self.stream_send:
            await self.stream_receive.send({"type": "lifespan.shutdown"})
            message = await receive()
            assert message["type"] in (
                "lifespan.shutdown.complete",
                "lifespan.shutdown.failed",
            )
            if message["type"] == "lifespan.shutdown.failed":
                await receive()



from __future__ import annotations

import asyncio
import functools
import hashlib
from concurrent import futures
from concurrent.futures import Future
from typing import Any, Awaitable, Generator, Generic, Protocol, TypeVar

from lilya._internal._urls import reverse as reverse

T = TypeVar("T")


class SupportsAsyncClose(Protocol):
    async def close(self) -> None: ...  # pragma: no cover


SupportsAsyncCloseType = TypeVar(
    "SupportsAsyncCloseType", bound=SupportsAsyncClose, covariant=False
)


try:
    hashlib.md5(b"data", usedforsecurity=False)  # type: ignore[call-arg]

    def md5_hexdigest(data: bytes, *, usedforsecurity: bool = True) -> str:  # pragma: no cover
        return hashlib.md5(  # type: ignore[call-arg]
            data, usedforsecurity=usedforsecurity
        ).hexdigest()

except TypeError:  # pragma: no cover

    def md5_hexdigest(data: bytes, *, usedforsecurity: bool = True) -> str:
        return hashlib.md5(data).hexdigest()


def is_async_callable(obj: Any) -> bool:
    """
    Validates if a given object is an async callable or not.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )


def run_sync(async_function: Awaitable) -> Any:
    """
    Runs the queries in sync mode
    """
    try:
        return asyncio.run(async_function)  # type: ignore
    except RuntimeError:
        with futures.ThreadPoolExecutor(max_workers=1) as executor:
            future: Future = executor.submit(asyncio.run, async_function)  # type: ignore
            return future.result()


class AsyncResourceHandler(Generic[SupportsAsyncCloseType]):
    """
    An asynchronous resource handler that acts as either an awaitable or a context manager.
    """

    __slots__ = ("awaitable_resource", "entered_resource")

    def __init__(self, awaitable_resource: Awaitable[SupportsAsyncCloseType]) -> None:
        """
        Initialize the AsyncResourceHandler with the provided awaitable resource.

        Args:
            awaitable_resource (Awaitable[SupportsAsyncCloseType]): The awaitable resource.
        """
        self.awaitable_resource = awaitable_resource

    def __await__(self) -> Generator[Any, None, SupportsAsyncCloseType]:
        """
        Allow the instance to be used as an awaitable.

        Returns:
            Generator[Any, None, SupportsAsyncCloseType]: The awaited resource.
        """
        return self.awaitable_resource.__await__()

    async def __aenter__(self) -> SupportsAsyncCloseType:
        """
        Enter the asynchronous context and obtain the resource.

        Returns:
            SupportsAsyncCloseType: The obtained resource.
        """
        self.entered_resource = await self.awaitable_resource
        return self.entered_resource

    async def __aexit__(self, *args: Any) -> None | bool:
        """
        Exit the asynchronous context and close the resource.

        Args:
            *args: Additional arguments.

        Returns:
            Union[None, bool]: Returns None or a boolean value indicating the exit status.
        """
        await self.entered_resource.close()
        return None



LILYA_DISCOVER_APP = "LILYA_DEFAULT_APP"
DEFAULT_TEMPLATE_NAME = "default"
APP_PARAMETER = "--app"
HELP_PARAMETER = "--help"
EXCLUDED_DIRECTIVES = ["createproject", "createapp", "createdeployment"]
IGNORE_DIRECTIVES = ["directives"]
DISCOVERY_FILES = ["application.py", "app.py", "main.py"]
DISCOVERY_FUNCTIONS = ["get_application", "get_app"]
TREAT_AS_PROJECT_DIRECTIVE = ["deployment"]



import sys

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

from lilya.exceptions import NotAuthorized
from lilya.protocols.permissions import PermissionProtocol
from lilya.types import ASGIApp, Receive, Scope, Send

P = ParamSpec("P")


class DenyAll(PermissionProtocol):
    def __init__(self, app: ASGIApp, *args: P.args, **kwargs: P.kwargs):
        super().__init__(app, *args, **kwargs)
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        raise NotAuthorized()



try:
    import click  # noqa
    import rich  # noqa
except ModuleNotFoundError:
    raise RuntimeError(
        "To use 'lilya cli' you need to install the dependencies. "
        "You can install them all by running 'pip install lilya[cli]'."
    ) from None
from lilya.cli.cli import lilya_cli


def run_cli() -> None:
    lilya_cli()


if __name__ == "__main__":  # pragma: no cover
    lilya_cli()



import copy
import operator
from typing import Any, Callable, TypeVar

empty = object()
RT = TypeVar("RT")  # return type


def new_method_proxy(func: Callable[..., RT]) -> Callable[..., RT]:
    def inner(self, *args: Any) -> RT:  # type: ignore
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    inner._mask_wrapped = False
    return inner


class LazyObject:  # pragma: no cover
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class.
    By subclassing, you have the opportunity to intercept and alter the
    instantiation. If you don't need to do that, use SimpleLazyObject.

    Based on Django and this article: https://coderbook.com/python/2020/04/23/how-to-make-lazy-python.html
    """

    _wrapped = None

    def __init__(self) -> None:
        self._wrapped = empty

    def __getattribute__(self, name: str) -> Any:
        if name == "_wrapped":
            return super().__getattribute__(name)
        value: Any = super().__getattribute__(name)
        if not getattr(value, "_mask_wrapped", True):
            raise AttributeError
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_wrapped":
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name: str) -> None:
        if name == "_wrapped":
            raise TypeError("Can't delete _wrapped.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self) -> Any:
        """
        Must be implemented by subclasses to initialize the wrapped object.
        """
        raise NotImplementedError("subclasses of LazyObject must provide a _setup() method")

    def __reduce__(self) -> Any:
        if self._wrapped is empty:
            self._setup()
        return (unpickle_lazyobject, (self._wrapped,))

    def __copy__(self) -> Any:
        if self._wrapped is empty:
            return type(self)()
        else:
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo: Any) -> Any:
        if self._wrapped is empty:
            result = type(self)()
            memo[id(self)] = result
            return result

    __getattr__ = new_method_proxy(getattr)
    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))  # type: ignore
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __gt__ = new_method_proxy(operator.gt)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)


def unpickle_lazyobject(wrapped: Any) -> Any:  # pragma: no cover
    """
    Used to unpickle lazy objects. Just return its argument, which will be the
    wrapped object.
    """
    return wrapped


class SimpleLazyObject(LazyObject):  # pragma: no cover
    """
    A lazy object initialized from any function.
    Designed for compound objects of unknown type.
    """

    def __init__(self, func: Callable[..., RT]) -> None:
        """
        Pass in a callable that returns the object to be wrapped.
        If copies are made of the resulting SimpleLazyObject, which can happen
        in various circumstances, then you must ensure that the
        callable can be safely run more than once and will return the same
        value.
        """
        self.__dict__["_setupfunc"] = func
        super().__init__()

    def _setup(self) -> Any:
        self._wrapped = self._setupfunc()

    def __repr__(self) -> str:
        if self._wrapped is empty:
            repr_attr = self._setupfunc
        else:
            repr_attr = self._wrapped
        return f"<{type(self).__name__}: {repr_attr!r}>"

    def __copy__(self) -> Any:
        if self._wrapped is empty:
            return SimpleLazyObject(self._setupfunc)
        else:
            # If initialized, return a copy of the wrapped object.
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo: Any) -> Any:
        if self._wrapped is empty:
            result = SimpleLazyObject(self._setupfunc)
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __add__ = new_method_proxy(operator.add)

    @new_method_proxy
    def __radd__(self, other: Any) -> Any:
        return other + self



from __future__ import annotations

from collections import deque
from dataclasses import asdict, is_dataclass
from enum import Enum
from pathlib import PurePath
from types import GeneratorType
from typing import Any, Generic, TypeVar, cast

from lilya._utils import is_class_and_subclass

T = TypeVar("T")


class Encoder(Generic[T]):
    """
    The base class for any custom encoder
    added to the system.
    """

    __type__: Any | None = None

    def is_type(self, value: Any) -> bool:
        return isinstance(value, self.__type__)

    def serialize(self, obj: T) -> T:
        raise NotImplementedError()


class DataclassEncoder(Encoder[Any]):

    def is_type(self, value: Any) -> bool:
        return cast(bool, is_dataclass(value))

    def serialize(self, obj: Any) -> Any:
        return asdict(obj)


class EnumEncoder(Encoder[Enum]):
    __type__ = Enum

    def serialize(self, obj: T) -> Any:
        return obj.value


class PurePathEncoder(Encoder[Any]):
    __type__ = PurePath

    def serialize(self, obj: PurePath) -> str:
        return str(obj)


class PrimitiveEncoder(Encoder[Any]):
    __type__ = (str, int, float, type(None))

    def serialize(self, obj: Any) -> Any:
        return obj


class DictEncoder(Encoder[dict]):
    __type__ = dict

    def serialize(self, obj: dict) -> dict:
        return obj


class StructureEncoder(Encoder[Any]):
    __type__ = (list, set, frozenset, GeneratorType, tuple, deque)

    def serialize(self, obj: Any) -> Any:
        serialized_objects = []
        for item in obj:
            for encoder in ENCODER_TYPES:
                if not encoder.is_type(item):
                    continue
                serialized_objects.append(encoder.serialize(item))
        return serialized_objects


ENCODER_TYPES: set[Encoder] = {
    DataclassEncoder(),
    EnumEncoder(),
    PurePathEncoder(),
    PrimitiveEncoder(),
    DictEncoder(),
    StructureEncoder(),
}


def register_encoder(encoder: Encoder[Any] | type[Encoder[Any]]) -> None:
    if is_class_and_subclass(encoder, Encoder):
        encoder = encoder()  # type: ignore
    ENCODER_TYPES.add(cast(Encoder[Any], encoder))


def json_encoder(value: Any) -> Any:
    result: Any = None

    for encoder in ENCODER_TYPES:
        try:
            if not encoder.is_type(value):
                continue
            result = encoder.serialize(value)
            break
        except (TypeError, AttributeError):
            continue

    if result is None:
        raise ValueError(f"object of type {type(value)} is not json serializable.")

    return result



from __future__ import annotations

import sys

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

from typing import runtime_checkable

from typing_extensions import Protocol

from lilya.types import ASGIApp, Receive, Scope, Send

P = ParamSpec("P")


@runtime_checkable
class PermissionProtocol(Protocol[P]):  # pragma: no cover
    __slots__ = ("app",)

    def __init__(self, app: ASGIApp, *args: P.args, **kwargs: P.kwargs): ...

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None: ...



from __future__ import annotations

import functools
import http.cookies
import json
import os
import stat
import typing
from datetime import datetime
from email.utils import format_datetime, formatdate
from mimetypes import guess_type
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Iterable,
    Literal,
    Mapping,
    NoReturn,
    Sequence,
    Union,
)
from urllib.parse import quote

import anyio

from lilya import status
from lilya._internal._encoders import json_encoder
from lilya._internal._helpers import HeaderHelper
from lilya.background import Task
from lilya.compat import md5_hexdigest
from lilya.concurrency import iterate_in_threadpool
from lilya.datastructures import URL, Header
from lilya.encoders import Encoder, register_encoder
from lilya.enums import Event, HTTPMethod, MediaType
from lilya.types import Receive, Scope, Send

Content = Union[str, bytes]
SyncContentStream = Iterable[Content]
AsyncContentStream = AsyncIterable[Content]
ContentStream = Union[AsyncContentStream, SyncContentStream]


class Response:
    media_type: str | None = None
    status_code: int | None = None
    charset: str = "utf-8"

    def __init__(
        self,
        content: Any = None,
        status_code: int = status.HTTP_200_OK,
        headers: Mapping[str, str] | None = None,
        cookies: Mapping[str, str] | Any | None = None,
        media_type: str | None = None,
        background: Task | None = None,
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ) -> None:
        if status_code is not None:
            self.status_code = status_code
        if media_type is not None:
            self.media_type = media_type
        self.background = background
        self.cookies = cookies
        self.encoders = encoders or []

        for encoder in self.encoders:
            register_encoder(encoder)

        self.body = self.make_response(content)
        self.raw_headers: list[Any] = []
        self.make_headers(headers)

    def make_response(self, content: Any) -> bytes | str:
        """
        Makes the Response object type.
        """
        if content is None or content is NoReturn or not content:
            return b""
        if isinstance(content, bytes):
            return content
        return content.encode(self.charset)  # type: ignore

    def make_headers(
        self, content_headers: Mapping[str, str] | dict[str, str] | None = None
    ) -> None:
        """
        Initializes the headers based on RFC specifications by setting appropriate conditions and restrictions.

        Args:
            content_headers (Union[Mapping[str, str], Dict[str, str], None], optional): Additional headers to include (default is None).
        """
        headers: dict[str, str] = {} if content_headers is None else content_headers  # type: ignore

        if HeaderHelper.has_entity_header_status(self.status_code):
            headers = HeaderHelper.remove_entity_headers(headers)
        if HeaderHelper.has_body_message(self.status_code):
            content_type = HeaderHelper.get_content_type(
                charset=self.charset, media_type=self.media_type
            )
            if hasattr(self, "body") and self.body is not None:
                headers.setdefault("content-length", str(len(self.body)))

            # Populates the content type if exists
            if content_type is not None:
                headers.setdefault("content-type", content_type)

        raw_headers = [
            (name.encode("latin-1"), f"{value}".encode(errors="surrogateescape"))
            for name, value in headers.items()
        ]
        self.raw_headers = raw_headers

    @property
    def headers(self) -> Header:
        if not hasattr(self, "_headers"):
            self._headers = Header(self.raw_headers)
        return self._headers

    def set_cookie(
        self,
        key: str,
        value: str = "",
        *,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        max_age: int | None = None,
        expires: datetime | str | int | None = None,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None:
        """
        Sets a cookie in the response headers.

        Args:
            key (str): The name of the cookie.
            value (str, optional): The value of the cookie.
            path (str, optional): The path for which the cookie is valid (default is '/').
            domain (Union[str, None], optional): The domain to which the cookie belongs.
            secure (bool, optional): If True, the cookie should only be sent over HTTPS.
            max_age (Union[int, None], optional): The maximum age of the cookie in seconds.
            expires (Union[Union[datetime, str, int], None], optional): The expiration date of the cookie.
            httponly (bool, optional): If True, the cookie should only be accessible through HTTP.
            samesite (Literal["lax", "strict", "none"], optional): SameSite attribute of the cookie.

        Raises:
            AssertionError: If samesite is not one of 'strict', 'lax', or 'none'.
        """
        cookie: http.cookies.BaseCookie[str] = http.cookies.SimpleCookie()
        cookie[key] = value
        if max_age is not None:
            cookie[key]["max-age"] = max_age
        if expires is not None:
            if isinstance(expires, datetime):
                cookie[key]["expires"] = format_datetime(expires, usegmt=True)
            else:
                cookie[key]["expires"] = expires
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be either 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.headers.add("set-cookie", cookie_val.encode("latin-1"))

    def delete_cookie(
        self,
        key: str,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None:
        """
        Deletes a cookie in the response headers by setting its max age and expiration to 0.

        Args:
            key (str): The name of the cookie to delete.
            path (str, optional): The path for which the cookie is valid (default is '/').
            domain (Union[str, None], optional): The domain to which the cookie belongs.
            secure (bool, optional): If True, the cookie should only be sent over HTTPS.
            httponly (bool, optional): If True, the cookie should only be accessible through HTTP.
            samesite (Literal["lax", "strict", "none"], optional): SameSite attribute of the cookie.
        """
        self.set_cookie(
            key,
            max_age=0,
            expires=0,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )

    @property
    def encoded_headers(self) -> list[Any]:
        headers: list[tuple[Any, Any]] = []

        for name, value in list(self.headers.multi_items()):
            if not isinstance(value, bytes):
                headers.append(
                    (name.encode("latin-1"), f"{value}".encode(errors="surrogateescape"))
                )
            else:
                headers.append((name.encode("latin-1"), value))
        return headers

    def message(self, prefix: str) -> dict[str, Any]:
        return {
            "type": prefix + "http.response.start",
            "status": self.status_code,
            "headers": self.encoded_headers,
        }

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        prefix = "websocket." if scope["type"] == "websocket" else ""
        await send(self.message(prefix=prefix))
        await send({"type": prefix + "http.response.body", "body": self.body})

        if self.background is not None:
            await self.background()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(media_type={self.media_type}, status_code={self.status_code}, charset={self.charset})"


class HTMLResponse(Response):
    media_type = MediaType.HTML


class Error(HTMLResponse):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR


class PlainText(Response):
    media_type = MediaType.TEXT


class JSONResponse(Response):
    media_type = MediaType.JSON

    def __init__(
        self,
        content: Any,
        status_code: int = status.HTTP_200_OK,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: Task | None = None,
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ) -> None:
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
            encoders=encoders,
        )

    def make_response(self, content: Any) -> bytes:
        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
        ).encode(self.charset)


class Ok(JSONResponse):
    media_type = MediaType.JSON


class RedirectResponse(Response):
    def __init__(
        self,
        url: str | URL,
        status_code: int = status.HTTP_307_TEMPORARY_REDIRECT,
        headers: Mapping[str, str] | None = None,
        background: Task | None = None,
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ) -> None:
        super().__init__(
            content=b"",
            status_code=status_code,
            headers=headers,
            background=background,
            encoders=encoders,
        )
        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")


class StreamingResponse(Response):
    body_iterator: AsyncContentStream

    def __init__(
        self,
        content: ContentStream,
        status_code: int = status.HTTP_200_OK,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: Task | None = None,
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ) -> None:
        self.encoders = encoders or []
        for encoder in self.encoders:
            register_encoder(encoder)

        if isinstance(content, AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background
        self.make_headers(headers)

    async def wait_for_disconnect(self, receive: Receive) -> None:
        while True:
            message = await receive()
            if message["type"] == Event.HTTP_DISCONNECT:
                break

    async def stream(self, send: Send) -> None:
        await send(self.message(prefix=""))

        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with anyio.create_task_group() as task_group:

            async def wrap(func: Callable[[], Awaitable[None]]) -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, functools.partial(self.stream, send))
            await wrap(functools.partial(self.wait_for_disconnect, receive))

        if self.background is not None:
            await self.background()


class FileResponse(Response):
    chunk_size = 64 * 1024

    def __init__(
        self,
        path: str | os.PathLike[str],
        status_code: int = status.HTTP_200_OK,
        headers: typing.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: Task | None = None,
        filename: str | None = None,
        stat_result: os.stat_result | None = None,
        method: str | None = None,
        content_disposition_type: str = "attachment",
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ) -> None:
        self.path = path
        self.status_code = status_code
        self.filename = filename
        self.send_header_only = method is not None and method.upper() == HTTPMethod.HEAD
        if media_type is None:
            media_type = guess_type(filename or path)[0] or MediaType.TEXT
        self.media_type = media_type
        self.background = background

        self.encoders = encoders or []
        for encoder in self.encoders:
            register_encoder(encoder)

        self.make_headers(headers)

        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = (
                    f"{content_disposition_type}; filename*=utf-8''{content_disposition_filename}"
                )
            else:
                content_disposition = f'{content_disposition_type}; filename="{self.filename}"'
            self.headers.setdefault("content-disposition", content_disposition)
        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result)

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)

        self.headers.setdefault("content-length", content_length)
        self.headers.setdefault("last-modified", last_modified)
        self.headers.setdefault("etag", etag)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self.stat_result is None:
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, self.path)
                self.set_stat_headers(stat_result)
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.") from None
            else:
                mode = stat_result.st_mode
                if not stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.") from None

        prefix = "websocket." if scope["type"] == "websocket" else ""
        await send(self.message(prefix=prefix))

        if self.send_header_only:
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        else:
            async with await anyio.open_file(self.path, mode="rb") as file:
                more_body = True
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    more_body = len(chunk) == self.chunk_size
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )
        if self.background is not None:
            await self.background()


class TemplateResponse(HTMLResponse):
    def __init__(
        self,
        template: Any,
        status_code: int = status.HTTP_200_OK,
        context: dict[str, Any] | None = None,
        background: Task | None = None,
        headers: dict[str, Any] | None = None,
        media_type: MediaType | str = MediaType.HTML,
        encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
    ):
        self.template = template
        self.context = context or {}
        content = self.template.render(context)
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
            encoders=encoders,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.context.get("request", {})
        extensions = request.get("extensions", {})
        if "http.response.debug" in extensions:
            await send(
                {
                    "type": "http.response.debug",
                    "info": {"template": self.template, "context": self.context},
                }
            )
        await super().__call__(scope, receive, send)


def make_response(
    content: Any,
    response_class: type[Response] = JSONResponse,
    status_code: int = status.HTTP_200_OK,
    headers: Mapping[str, str] | None = None,
    background: Task | None = None,
    encoders: Sequence[Encoder[Any]] | Sequence[type[Encoder[Any]]] | None = None,
) -> Response:
    """
    Build JSON responses from a given content and
    providing extra parameters.
    """
    app = json_encoder(content) if content is not None else None

    return response_class(
        content=app,
        status_code=status_code,
        headers=headers,
        media_type=MediaType.JSON,
        background=background,
        encoders=encoders,
    )



import functools
from typing import Any, AsyncIterator, Awaitable, Callable, Iterable, TypeVar

import anyio.to_thread
from anyio import create_task_group, get_cancelled_exc_class, to_thread

from lilya.compat import is_async_callable

T = TypeVar("T")


async def run_in_threadpool(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Runs a callable in a threapool.
    Make sure the callable is always async.
    """
    def_func = AsyncCallable(func)
    return await def_func(*args, **kwargs)


def enforce_async_callable(func: Callable[..., Any]) -> Callable[..., Awaitable[T]]:
    """
    Enforces the callable to be async by returning an AsyncCallable.
    """
    return func if is_async_callable(func) else AsyncCallable(func)  # type:ignore


class AsyncCallable:
    """
    Creates an async callable and when called, runs in a thread pool.
    """

    __slots__ = ("_callable", "default_kwargs")

    def __init__(self, func: Callable[..., Any], **kwargs: Any) -> None:
        self._callable = func
        self.default_kwargs = kwargs

    def __call__(self, *args: Any, **kwargs: Any) -> Awaitable[T]:
        combined_kwargs = {**self.default_kwargs, **kwargs}
        return anyio.to_thread.run_sync(
            functools.partial(self._callable, *args, **combined_kwargs)
        )

    async def run_in_threadpool(self, *args: Any, **kwargs: Any) -> T:
        return await self(*args, **kwargs)


async def iterate_in_threadpool(iterator: Iterable[T]) -> AsyncIterator[T]:
    async def worker() -> AsyncIterator[T]:
        for item in iterator:
            yield item

    async with create_task_group():
        stream = await to_thread.run_sync(worker)
        try:
            async for item in stream:
                yield item
        except get_cancelled_exc_class():
            ...



from __future__ import annotations

import copy
import warnings
from typing import TYPE_CHECKING, Any

from typing_extensions import Annotated, Doc

from lilya.datastructures import URL
from lilya.types import Scope

if TYPE_CHECKING:
    from lilya.requests import Request
    from lilya.routing import BasePath


class Context:
    """
    `Context` class is used for the handlers context of the
    scope of the call.

    When a `context` is passed through the handler,
    the context will be aware of the decoractor handler
    itself.

    You will probably not need the context or change it but it is here in
    case you decide to use it.

    !!! Tip
        The context only exists in the handlers and nothing else which you can also
        see it as `request context` sort of approach.

    **Example**

    ```python
    from lilya.apps import Lilya
    from lilya.requests import Request
    from lilya.context import Context


    async def home(context: Context) -> None:
        ...
    ```
    """

    def __init__(
        self,
        __handler__: Annotated[
            BasePath,
            Doc(
                """
                The [handler](https://lilya.dev/routing) where the context will be. placed.

                To avoid any adulteration of the the original route handler, the context performs
                a shallow copy of the original handler itself.
                """
            ),
        ],
        __request__: Annotated[
            Request,
            Doc(
                """
                A [Request](https://lilya.dev/references/request/) class object.
                """
            ),
        ],
    ) -> None:
        self.__handler__ = copy.copy(__handler__)
        self.__request__ = __request__

    @property
    def handler(self) -> BasePath:
        return self.__handler__

    @property
    def request(self) -> Request:
        return self.__request__

    @property
    def user(self) -> Any:
        return self.request.user

    @property
    def settings(self) -> Any:
        return self.request.app.settings

    @property
    def scope(self) -> Scope:
        return self.request.scope

    @property
    def app(self) -> Any:
        return self.request.scope["app"]

    def add_to_context(
        self,
        key: Annotated[
            Any,
            Doc(
                """
                The key value to be added to the context dictionary.
                """
            ),
        ],
        value: Annotated[
            Any,
            Doc(
                """
                The value value to be added to the context dictionary and map with the key.
                """
            ),
        ],
    ) -> None:
        """
        Adds a key to the context of an handler.

        This can be particularly useful if you are programatically
        building the handler or for any other specific and unique operation.

        **Example**

        ```python
        from typing import Any
        from lilya.context import Context

        async def get_data(context: Context) -> Any:
            context.update{"name": "Lilya"}
            return context.get_context_data()
        ```
        """
        if key in self.__dict__:
            warnings.warn(
                f"The key: '{key} already exists in context and it will be overritten.",
                stacklevel=2,
            )
        setattr(self, key, value)

    def get_context_data(self, **kwargs: Any) -> dict[Any, Any]:
        """
        Returns the context in a python dictionary like structure.
        """
        context_data = {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("__") and not k.endswith("__")
        }
        return context_data

    def path_for(
        self,
        name: Annotated[
            str,
            Doc(
                """
                The `name` given in a `Gateway` or `Include` objects.

                **Example**

                ```python
                from lilya import Lilya, Gateway

                Gateway(handler=..., name="view-users")
                ```
                """
            ),
        ],
        **path_params: Annotated[
            Any,
            Doc(
                """
                Any additional *path_params* declared in the URL of the
                handler and are needed to *reverse* the names and return
                the proper URL.
                """
            ),
        ],
    ) -> str:
        url: URL = self.request.path_for(name, **path_params)
        return str(url)



from __future__ import annotations

from typing import Mapping

from lilya import status


class HeaderHelper:
    # According to https://tools.ietf.org/html/rfc2616#section-7.1
    entity_headers: list[str] = [
        "allow",
        "content-encoding",
        "content-language",
        "content-length",
        "content-location",
        "content-md5",
        "content-range",
        "content-type",
        "expires",
        "last-modified",
        "extension-header",
    ]

    @classmethod
    def is_entity_header(cls, header: str) -> bool:
        return bool(header.lower() in cls.entity_headers)

    @classmethod
    def has_entity_header_status(cls, status_code: int) -> bool:
        return bool(
            status_code in (status.HTTP_304_NOT_MODIFIED, status.HTTP_412_PRECONDITION_FAILED)
        )

    @classmethod
    def remove_entity_headers(
        cls, headers: Mapping[str, str], allowed: tuple[str, str] | None = None
    ) -> dict[str, str]:
        """
        Removes all the entity headers present in the headers given.
        According to RFC 2616 Section 10.3.5.

        https://tools.ietf.org/html/rfc2616#section-10.3.5
        """
        if allowed is None:
            allowed = ("content-location", "expires")

        _allowed: set[str] = {h.lower() for h in allowed}
        headers = {
            header: value
            for header, value in headers.items()
            if not cls.is_entity_header(header) or header.lower() in _allowed
        }
        return headers

    @classmethod
    def has_body_message(cls, status_code: int) -> bool:
        """
        Based on the RFC specificiation the response status of 1XX, 204 and 304
        body and length **MUST NOT** be included.

        https://tools.ietf.org/html/rfc2616#section-4.4
        https://tools.ietf.org/html/rfc2616#section-4.3
        """
        return bool(status_code not in (204, 304) and not (100 <= status_code < 200))

    @classmethod
    def get_content_type(self, charset: str, media_type: str | None = None) -> str:
        """
        Builds the content-type based on the media type and charset.
        """
        if media_type is not None and media_type.startswith("text/"):
            return f"{media_type}; charset={charset}"
        return media_type



from __future__ import annotations

import math
import re
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, ClassVar, Generic, TypeVar

from lilya._utils import is_class_and_subclass

T = TypeVar("T")

# path_regex = re.compile(r"{(.*?)}")
path_regex = re.compile(r"{([^:]+):([^}]+)}")


@dataclass(frozen=True, eq=True)
class Transformer(Generic[T]):
    """
    Base for all path transformers
    of lilya. Dataclasses are used
    for the simplicity of the syntax
    """

    regex: ClassVar[str] = ""

    def transform(self, value: str) -> T:
        raise NotImplementedError()  # pragma: no cover

    def normalise(self, value: T) -> str:  # pragma: no cover
        raise NotImplementedError()  # pragma: no cover


@dataclass(frozen=True, eq=True)
class StringTransformer(Transformer[str]):
    regex = "[^/]+"

    def transform(self, value: str) -> str:
        return value

    def normalise(self, value: str) -> str:
        value = str(value)
        assert "/" not in value, "May not contain path separators"
        assert value, "Must not be empty"
        return value


@dataclass(frozen=True, eq=True)
class PathTransformer(Transformer[str]):
    regex = ".*"

    def transform(self, value: str) -> str:
        return str(value)

    def normalise(self, value: str) -> str:
        return str(value)


@dataclass(frozen=True, eq=True)
class IntegerTransformer(Transformer[int]):
    regex = "[0-9]+"

    def transform(self, value: str) -> int:
        return int(value)

    def normalise(self, value: int) -> str:
        value = int(value)
        assert value >= 0, "Negative integers are not supported"
        return str(value)


@dataclass(frozen=True, eq=True)
class FloatTransformer(Transformer[float]):
    regex = r"[0-9]+(\.[0-9]+)?"

    def transform(self, value: str) -> float:
        return float(value)

    def normalise(self, value: float) -> str:
        value = float(value)
        assert value >= 0.0, "Negative floats are not supported"
        assert not math.isnan(value), "NaN values are not supported"
        assert not math.isinf(value), "Infinite values are not supported"
        return (f"{value:0.20f}").rstrip("0").rstrip(".")


@dataclass(frozen=True, eq=True)
class UUIDTransformer(Transformer[uuid.UUID]):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def transform(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def normalise(self, value: uuid.UUID) -> str:
        return str(value)


@dataclass(frozen=True, eq=True)
class DatetimeTransformer(Transformer[datetime]):
    regex = "[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}(.[0-9]+)?"

    def transform(self, value: str) -> datetime:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")

    def normalise(self, value: datetime) -> str:
        return value.strftime("%Y-%m-%dT%H:%M:%S")


# Available converter types
TRANSFORMER_TYPES: dict[str, Transformer] = {
    "str": StringTransformer(),
    "path": PathTransformer(),
    "int": IntegerTransformer(),
    "float": FloatTransformer(),
    "uuid": UUIDTransformer(),
    "datetime": DatetimeTransformer(),
}

TRANSFORMER_PYTHON_TYPES = {v.__class__.__name__: k for k, v in TRANSFORMER_TYPES.items()}


def register_path_transformer(key: str, transformer: Transformer[Any]) -> None:
    """
    Adds custom transformers to the already known dictionary of path
    transformers.
    """
    if is_class_and_subclass(transformer, Transformer):
        transformer = transformer()  # type: ignore

    TRANSFORMER_TYPES[key] = transformer
    TRANSFORMER_PYTHON_TYPES[transformer.__class__.__name__] = key



from lilya._internal._path_transformers import (
    TRANSFORMER_TYPES,
    Transformer,
    register_path_transformer,
)

__all__ = [
    "TRANSFORMER_TYPES",
    "Transformer",
    "register_path_transformer",
]



from __future__ import annotations

import sys

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

from typing import Any, Callable, Sequence

import anyio

from lilya._internal import Repr
from lilya.concurrency import enforce_async_callable

P = ParamSpec("P")


class Task(Repr):
    """
    `Task` as a single instance can be easily achieved.

    **Example**

    ```python
    from datetime import datetime
    from typing import Dict

    from lilya.apps import Lilya
    from lilya.background import Task
    from lilya.requests import Request
    from lilya.responses import Response
    from lilya.routing import Path


    async def send_email_notification(email: str, message: str):
        send_notification(email, message)


    async def create_user(request: Request) -> Response(Dict[str, str]):
        return Response(
            {"message": "Email sent"},
            background=Task(
                send_email_notification,
                email="example@lilya.dev",
                message="Thank you for registering.",
            ),
        )


    app = Lilya(
        routes=[
            Path(
                "/register",
                create_user,
                methods=["POST"],
            )
        ]
    )
    ```
    """

    __slots__ = ("func", "args", "kwargs")

    def __init__(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        self.func = enforce_async_callable(func)
        self.args = args
        self.kwargs = kwargs

    async def __call__(self) -> None:
        await self.func(*self.args, **self.kwargs)


class Tasks(Task):
    """
    A container for background tasks.

    When `as_group` is set to `True`, it will run all the tasks
    concurrently (as a group).

    **Example**

    ```python
    from datetime import datetime

    from lilya.apps import Lilya
    from lilya.background import Tasks
    from lilya.requests import Request
    from lilya.responses import Response
    from lilya.routing import Path


    async def send_email_notification(email: str, message: str):
        send_notification(email, message)


    def write_in_file(email: str):
        with open("log.txt", mode="w") as log:
            now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            content = f"Notification sent @ {now} to: {email}"
            log.write(content)


    async def create_user(request: Request):
        background_tasks = Tasks(as_group=True)
        background_tasks.add_task(
            send_email_notification,
            email="example@lilya.dev",
            message="Thank you for registering.",
        )
        background_tasks.add_task(write_in_file, email=request.user.email)

        return Response({"message": "Email sent"}, background=background_tasks)


    app = Lilya(
        routes=[
            Path(
                "/register",
                create_user,
                methods=["POST"],
            )
        ]
    )
    ```
    """

    __slots__ = ("tasks", "as_group")

    def __init__(self, tasks: Sequence[Task] | None = None, as_group: bool = False):
        self.tasks = list(tasks) if tasks else []
        self.as_group = as_group

    def add_task(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        task = Task(func, *args, **kwargs)
        self.tasks.append(task)

    async def run_single(self) -> None:
        for task in self.tasks:
            await task()

    async def run_as_group(self) -> None:
        async with anyio.create_task_group() as group:
            for task in self.tasks:
                group.start_soon(task)

    async def __call__(self) -> None:
        if not self.as_group:
            await self.run_single()
        else:
            await self.run_as_group()



from __future__ import annotations

import os
import sys
from dataclasses import dataclass
from importlib import import_module
from pathlib import Path

from lilya.apps import ChildLilya, Lilya
from lilya.cli.constants import DISCOVERY_FILES, DISCOVERY_FUNCTIONS, LILYA_DISCOVER_APP
from lilya.cli.terminal import Print

printer = Print()


@dataclass
class Scaffold:
    """
    Simple Application scaffold that holds the
    information about the app and the path to
    the same app.
    """

    path: str
    app: Lilya | ChildLilya


@dataclass
class DirectiveEnv:
    """
    Loads an arbitraty application into the object
    and returns the App.
    """

    path: str | None = None
    app: Lilya | ChildLilya | None = None
    command_path: str | None = None

    def load_from_env(self, path: str | None = None) -> DirectiveEnv:
        """
        Loads the environment variables into the scaffold.
        """
        # Adds the current path where the command is being invoked
        # To the system path
        cwd = Path().cwd()
        command_path = str(cwd)
        if command_path not in sys.path:
            sys.path.append(command_path)
        try:
            import dotenv

            dotenv.load_dotenv()
        except ImportError:
            ...

        _path = os.getenv(LILYA_DISCOVER_APP) if not path else path
        _app = self.find_app(path=_path, cwd=cwd)

        return DirectiveEnv(path=_app.path, app=_app.app, command_path=command_path)

    def import_app_from_string(cls, path: str | None = None) -> Scaffold:
        if path is None:
            raise OSError(
                "Path cannot be None. Set env `LILYA_DEFAULT_APP` or use `--app` instead."
            )
        module_str_path, app_name = path.split(":")
        module = import_module(module_str_path)
        app = getattr(module, app_name)
        return Scaffold(path=path, app=app)

    def _get_folders(self, path: Path) -> list[str]:
        """
        Lists all the folders and checks if there is any file from the DISCOVERY_FILES available
        """
        return [directory.path for directory in os.scandir(path) if directory.is_dir()]

    def _find_app_in_folder(self, path: Path, cwd: Path) -> Scaffold | None:
        """
        Iterates inside the folder and looks up to the DISCOVERY_FILES.
        """
        for discovery_file in DISCOVERY_FILES:
            filename = f"{str(path)}/{discovery_file}"
            if not os.path.exists(filename):
                continue

            file_path = path / discovery_file
            dotted_path = ".".join(file_path.relative_to(cwd).with_suffix("").parts)

            # Load file from module
            module = import_module(dotted_path)

            # Iterates through the elements of the module.
            for attr, value in module.__dict__.items():
                if isinstance(value, Lilya):
                    app_path = f"{dotted_path}:{attr}"
                    return Scaffold(app=value, path=app_path)

            # Iterate over default pattern application functions
            for func in DISCOVERY_FUNCTIONS:
                if hasattr(module, func):
                    app_path = f"{dotted_path}:{func}"
                    fn = getattr(module, func)
                    return Scaffold(app=fn(), path=app_path)
        return None

    def find_app(self, path: str | None, cwd: Path) -> Scaffold:
        """
        Loads the application based on the path provided via env var.

        If no --app is provided, goes into auto discovery up to one level.
        """

        if path:
            return self.import_app_from_string(path)

        scaffold: Scaffold | None = None

        # Check current folder
        scaffold = self._find_app_in_folder(cwd, cwd)
        if scaffold:
            return scaffold

        # Goes into auto discovery mode for one level, only.
        folders = self._get_folders(cwd)

        for folder in folders:
            folder_path = cwd / folder
            scaffold = self._find_app_in_folder(folder_path, cwd)

            if not scaffold:
                continue
            break

        if not scaffold:
            raise OSError("Could not find a Lilya application.")
        return scaffold



from __future__ import annotations

from typing import cast

from lilya._internal._module_loading import import_string

try:
    from a2wsgi.wsgi import WSGIMiddleware as A2WSGIMiddleware  # noqa
    from a2wsgi.wsgi_typing import WSGIApp
except ModuleNotFoundError:
    raise RuntimeError(
        "You need to install the package `a2wsgi` to be able to use this middleware. "
        "Simply run `pip install a2wsgi`."
    ) from None


class WSGIMiddleware(A2WSGIMiddleware):
    def __init__(self, app: WSGIApp | str, workers: int = 10) -> None:
        if isinstance(app, str):
            app = cast(WSGIApp, import_string(app))
        super().__init__(app, workers)
